# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os.path

from ..decorators import *
from .AnnotatedType import AnnotatedType
from .AttributeGroupType import AttributeGroupType
from .AttributeType import AttributeType
from .BooleanType import BooleanType
from .ChoiceElement import ChoiceElement
from .ComplexContentElement import ComplexContentElement
from .GroupType import GroupType
from .NCNameType import NCNameType
from .SimpleContentElement import SimpleContentElement
from .WildcardType import WildcardType

logger = logging.getLogger(__name__)

@attribute(local_name='name', type=NCNameType)
@attribute(local_name='mixed', type=BooleanType, default=False)
@attribute(local_name='abstract', type=BooleanType, default=False)
@attribute(local_name='final', enum=['#all', 'extension', 'restriction'])
@attribute(local_name='block', enum=['#all', 'extension', 'restriction'])
@attribute(local_name='*', )
@element(local_name='simpleContent', list='tags', cls=SimpleContentElement, min=0, max=None)
@element(local_name='complexContent', list='tags', cls=ComplexContentElement, min=0, max=None)
@element(local_name='group', list='tags', cls=GroupType, min=0)
@element(local_name='all', list='tags',
    cls=('expatriate.model.xs.AllType', 'AllType'), min=0)
@element(local_name='choice', list='tags', cls=ChoiceElement, min=0)
@element(local_name='sequence', list='tags', cls=GroupType, min=0)
@element(local_name='attribute', list='tags', cls=AttributeType, min=0, max=None)
@element(local_name='attributeGroup', list='tags', cls=AttributeGroupType, min=0, max=None)
@element(local_name='anyAttribute', list='tags', cls=WildcardType, min=0)
class ComplexTypeType(AnnotatedType):
    # TODO .mixed & simpleContent sub-elements are mutulally exclusive

    def stub(self, path, schema):
        class_name = ''.join(cap_first(w) for w in self.name.split('_'))
        if not class_name.endswith('Type'):
            class_name = class_name + 'Type'

        super().stub(path, schema, class_name)



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from ..decorators import *
from .AllNniType import AllNniType
from .NonNegativeIntegerType import NonNegativeIntegerType
from .WildcardType import WildcardType

logger = logging.getLogger(__name__)

@attribute(local_name='minOccurs', type=NonNegativeIntegerType, default=1)
@attribute(local_name='maxOccurs', type=AllNniType, default=1)
class AnyElement(WildcardType):
    def get_defs(self, schema):
        model_map = {'elements': [], 'attributes': {}}

        tag = {'tag_name': '*', 'min': 0}
        if self.namespace != '##any':
            tag['namespace'] = self.namespace
        model_map['elements'].append(tag)

        return model_map



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from ..decorators import *
from .AnnotationElement import AnnotationElement
from .AnyTypeType import AnyTypeType
from .AnyUriType import AnyUriType
from .AttributeGroupType import AttributeGroupType
from .ComplexTypeType import ComplexTypeType
from .GroupType import GroupType
from .IdType import IdType
from .SimpleTypeType import SimpleTypeType

logger = logging.getLogger(__name__)

@attribute(local_name='schemaLocation', type=AnyUriType, required=True)
@attribute(local_name='id', type=IdType)
@element(local_name='annotation', list='tags', cls=AnnotationElement, min=0, max=None)
@element(local_name='simpleType', list='tags', cls=SimpleTypeType, min=0, max=None)
@element(local_name='complexType', list='tags', cls=ComplexTypeType, min=0, max=None)
@element(local_name='group', list='tags', cls=GroupType, min=0, max=None)
@element(local_name='attributeGroup', list='tags', cls=AttributeGroupType, min=0, max=None)
class RedefineElement(AnyTypeType):
    pass



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re

from ..decorators import *
from .AnySimpleType import AnySimpleType

logger = logging.getLogger(__name__)

class StringType(AnySimpleType):
    def get_value_pattern(self):
        return None

    def get_value_enum(self):
        return None

    def parse_value(self, value):
        if not isinstance(value, str):
            raise TypeError('xs:string requires a str value for initialization, got ' + value.__class__.__name__)

        p = self.get_value_pattern()
        if p is not None and not re.fullmatch(p, value):
            raise ValueError(self.__class__.__name__ + ' requires a value matching ' + p)

        e = self.get_value_enum()
        if e is not None and value not in e:
            raise ValueError(self.__class__.__name__ + ' requires a value in ' + str(e))

        return value



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .Function import Function
from ..exceptions import XPathSyntaxException

logger = logging.getLogger(__name__)
class Predicate(object):
    def __init__(self):
        self.children = []

    def evaluate(self, context_node, context_position, context_size, variables):
        if len(self.children) != 1:
            raise XPathSyntaxException('Predicate can only have 1 expression')

        v = self.children[0].evaluate(context_node, context_position, context_size, variables)

        if isinstance(v, bool):
            logger.debug('Boolean predicate subexpression: ' + str(v))
            return v
        elif isinstance(v, int) or isinstance(v, float):
            logger.debug('Numeric result for predicate subexpression: ' + str(v) + '; comparing to position()')
            return Function.f_position((), context_node, context_position, context_size, variables) == v
        else:
            v_b = Function.f_boolean((v,), context_node, context_position, context_size, variables)
            logger.debug('Converting predicate subexpression result ' + str(v) + ' to boolean: '  + str(v_b))
            return v_b

    def __str__(self):
        return 'Predicate ' + hex(id(self)) + ': [' + ','.join([str(x) for x in self.children]) + ']'



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import logging

from .exceptions import *
from .Mapper import Mapper

logger = logging.getLogger(__name__)

class AttributeMapper(Mapper):
    '''
        Decorator to map xml elements to model children

        **kwargs**

        namespace
            The xml namespace to match. It can also be * to match any namespace.
            If not specified, it defaults to the parent element's namespace.
        local_name
            Required. The local name of the xml attribute we're matching. It can
            also be * to match any local name.
        into
            The python attribute to store the value of the attribute into.
            Defaults to the local_name if not specified.

        type
            The type of the expected value. Types are stored directly as data,
            no enclosed in a model class. Types usually restrict the domain
            values.
        enum
            Enumeration the attribute's value must be from
        pattern
            Pattern which the value of the attribute must match.

        required
            True or False (default). Specifies if the attribute is required.
        default
            The default value of an attribute if it isn't specified. The
            (implicit) default is None or the first item of an *enum*.

        min
            The minimum value of the attribute. Can be numeric or None (the
            default).
        max
            The maximum value of the attribute. Can be numeric or None (the
            default).

        prohibited
            The attribute should not appear in the element.
    '''
    def __init__(self, **kwargs):
        if 'local_name' not in kwargs:
            raise DecoratorException('Attributes need at least local_name defined')

        super().__init__(**kwargs)

    def get_attr_name(self):
        if 'into' in self._kwargs:
            return self._kwargs['into']
        else:
            return self._kwargs['local_name'].replace('-', '_')

    def initialize(self, model):
        attr_name = self.get_attr_name()

        if 'default' in self._kwargs:
            default_value = self._kwargs['default']
        else:
            default_value = None

        setattr(model, attr_name, default_value)

        logger.debug('Initialized ' + str(model) + ' attribute ' + attr_name
            + ' to default ' + str(default_value))

    def get_namespace(self):
        if 'namespace' in self._kwargs:
            return self._kwargs['namespace']
        else:
            return None

    def get_local_name(self):
        return self._kwargs['local_name']

    def matches(self, attr, model):
        from .Model import Model

        matches = (self.get_namespace(), self.get_local_name()) in (
            (attr.namespace, attr.local_name),
            (None, attr.local_name),
            (attr.namespace, Model.ANY_LOCAL_NAME),
            (Model.ANY_NAMESPACE, Model.ANY_LOCAL_NAME)
        )

        if matches:
            logger.debug('AttributeMapper matches ' + str(attr))
        else:
            logger.debug('AttributeMapper does not match ' + str(attr))

        return matches

    def parse_in(self, model, attr):
        logger.debug('Parsing attribute ' + attr.name + ' using kwargs: ' + str(self._kwargs))

        name = self.get_attr_name()
        value = attr.value

        if 'enum' in self._kwargs and value not in self._kwargs['enum']:
            raise EnumerationException(name + ' attribute must be one of '
                + str(self._kwargs['enum']) + ': ' + str(value))

        # convert value
        if 'type' in self._kwargs:
            logger.debug('Parsing ' + str(value) + ' as '
                + str(self._kwargs['type']))
            type_ = self._kwargs['type']
            # load from a tuple (module_name, class_name) to defer class load
            if isinstance(type_, tuple):
                mod = importlib.import_module(type_[0])
                type_ = getattr(mod, type_[1])
            value = type_().parse_value(value)

        logger.debug('Parsed attribute ' + name + ' = ' + str(value))

        setattr(model, name, value)

    def validate(self, model):
        name = self.get_attr_name()
        logger.debug('Validating attribute ' + str(self._kwargs))

        name = self._kwargs['local_name']

        # check that required attributes are defined
        if (
            'required' in self._kwargs
            and self._kwargs['required']
            and (
                not hasattr(model, name)
                or getattr(model, name) is None
            )
        ):
            raise RequiredAttributeException(str(model) + ' must define ' + name + ' attribute')

        # check that prohibited attributes are not defined
        if (
            'prohibited' in self._kwargs
            and self._kwargs['prohibited']
            and hasattr(model, name)
        ):
            raise ProhibitedAttributeException(str(model) + ' must not define ' + name + ' attribute')

    def produce_in(self, el, model):
        from .Model import Model
        from .xs.StringType import StringType

        if 'prohibited' in self._kwargs and self._kwargs['prohibited']:
            return

        name = self.get_attr_name()
        value = getattr(model, name)

        if value is None:
            return
        if 'default' in self._kwargs and value == self._kwargs['default']:
            return

        logger.debug(str(self) + ' producing ' + str(model) + ' attribute '
            + name + ' according to ' + str(self._kwargs))

        if 'namespace' in self._kwargs:
            namespace = self._kwargs['namespace']
        else:
            namespace = el.namespace

        prefix = el.namespace_to_prefix(namespace)

        if self._kwargs['local_name'] == Model.ANY_LOCAL_NAME:
            return
        else:
            local_name = self._kwargs['local_name']

        # if model's namespace doesn't match attribute's, then we need to include it
        if 'namespace' in self._kwargs and self._kwargs['namespace'] != el.namespace:
            attr_name = prefix + ':' + local_name
        else:
            attr_name = local_name

        if 'type' in self._kwargs:
            logger.debug(str(model) + ' Producing ' + str(value) + ' as '
                + str(self._kwargs['type']) + ' type')
            type_ = self._kwargs['type']()
            v = type_.produce_value(value)

            el.attributes[attr_name] = v

        elif 'enum' in self._kwargs:
            if value not in self._kwargs['enum']:
                raise EnumerationException(str(model) + '.' + name
                    + ' attribute must be one of ' + str(self._kwargs['enum'])
                    + ': ' + str(value))
            el.attributes[attr_name] = value

        else:
            # otherwise, we default to producing as string
            logger.debug(str(model) + ' Producing ' + str(value)
                + ' as String type')
            type_ = StringType()
            v = type_.produce_value(value)
            el.attributes[attr_name] = v



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re

from ..decorators import *
from .StringType import StringType

logger = logging.getLogger(__name__)

class List(StringType):
    # abstract
    def parse_value(self, value):
        value = super().parse_value(value)

        if len(value) < 1:
            raise ValueError('xs:List must contain at least 1 character')

        r = []
        for i in re.split(r'\s+', value):
            r.append(self.parse_item(i))

        return tuple(r)

    def produce_value(self, value):
        r = []
        for i in value:
            r.append(self.produce_item(i))
        return ' '.join(r)

    def parse_item(self, item_value):
        import inspect
        raise NotImplementedError(inspect.stack()[0][3] + '() has not been implemented in subclass: ' + self.__class__.__name__)

    def produce_item(self, item_value):
        import inspect
        raise NotImplementedError(inspect.stack()[0][3] + '() has not been implemented in subclass: ' + self.__class__.__name__)



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from ..decorators import *
from .AllNniType import AllNniType
from .AnnotatedType import AnnotatedType
from .AnyElement import AnyElement
from .NCNameType import NCNameType
from .NonNegativeIntegerType import NonNegativeIntegerType
from .QNameType import QNameType

logger = logging.getLogger(__name__)

@attribute(local_name='name', type=NCNameType)
@attribute(local_name='ref', type=QNameType)
@attribute(local_name='minOccurs', type=NonNegativeIntegerType, default=1)
@attribute(local_name='maxOccurs', type=AllNniType, default=1)
@attribute(local_name='*', )
@element(local_name='any', list='tags', cls=AnyElement, min=0, max=None)
@element(local_name='element', list='tags',
    cls=('expatriate.model.xs.ElementType', 'ElementType'),
        min=0, max=None)
@element(local_name='group', list='tags',
    cls=('expatriate.model.xs.GroupType', 'GroupType'),
    min=0, max=None)
@element(local_name='all', list='tags',
    cls=('expatriate.model.xs.AllType', 'AllType'),
    min=0, max=None)
@element(local_name='choice', list='tags',
    cls=('expatriate.model.xs.ChoiceElement', 'ChoiceElement'),
    min=0, max=None)
@element(local_name='sequence', list='tags',
    cls=('expatriate.model.xs.GroupType', 'GroupType'),
    min=0, max=None)
class GroupType(AnnotatedType):
    pass



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging
import math
import re

from ..exceptions import XPathSyntaxException

logger = logging.getLogger(__name__)
class Function(object):
    # Node Set Functions
    def f_last(args, context_node, context_position, context_size, variables):
        if len(args) != 0:
            raise XPathSyntaxException('last() expects 0 arguments')

        return context_size

    def f_position(args, context_node, context_position, context_size, variables):
        if len(args) != 0:
            raise XPathSyntaxException('position() expects 0 arguments')

        return context_position

    def f_count(args, context_node, context_position, context_size, variables):
        if len(args) != 1:
            raise XPathSyntaxException('count() expects 1 argument')

        return len(args[0])

    def f_id(args, context_node, context_position, context_size, variables):
        if len(args) != 1:
            raise XPathSyntaxException('id() expects 1 argument')

        from ..Node import Node
        ids = []
        if isinstance(args[0], list): # node list
            for n in args[0]:
                if not isinstance(n, Node):
                    raise XPathSyntaxException('Cannot determine the string value of ' + str(n))
                s = n.get_string_value()
                logger.debug('Using ' + str(s) + ' as string-value of ' + str(n))
                ids.extend(re.split(r'[\x20\x09\x0D\x0A]+', s))
        else:
            ids = re.split(r'[\x20\x09\x0D\x0A]+', Function.f_string((args[0],), context_node, context_position, context_size, variables))

        doc = context_node.get_document()
        ns = []
        for i in ids:
            try:
                ns.append(doc.find_by_id(i))
            except KeyError:
                logger.debug('Could not find element with id: ' + str(i))
                pass

        if len(ns) == 1:
            return ns[0]
        else:
            return ns

    def f_local_name(args, context_node, context_position, context_size, variables):
        if len(args) == 0:
            a = [context_node]
        elif len(args) == 1:
            a = args[0]
        else:
            raise XPathSyntaxException('local-name() expects 1 argument or none')

        if len(a) <= 0:
            return ''

        from ..Document import Document
        first_node = Document.ordered_first(a)

        if not hasattr(first_node, 'get_expanded_name'):
            return ''
        else:
            return first_node.get_expanded_name()[1]

    def f_namespace_uri(args, context_node, context_position, context_size, variables):
        if len(args) == 0:
            a = [context_node]
        elif len(args) == 1:
            a = args[0]
        else:
            raise XPathSyntaxException('namespace-uri() expects 1 argument or none')

        if len(a) <= 0:
            return ''

        from ..Document import Document
        first_node = Document.ordered_first(a)

        if not hasattr(first_node, 'namespace'):
            return ''
        else:
            return first_node.namespace

    def f_name(args, context_node, context_position, context_size, variables):
        if len(args) == 0:
            a = [context_node]
        elif len(args) == 1:
            a = args[0]
        else:
            raise XPathSyntaxException('name() expects 1 argument or none')

        if len(a) <= 0:
            return ''

        from ..Document import Document
        first_node = Document.ordered_first(a)

        if not hasattr(first_node, 'name'):
            return ''
        else:
            return first_node.name

    # String Functions

    def f_string(args, context_node, context_position, context_size, variables):
        from ..Node import Node
        if len(args) == 0:
            a = [context_node]
        elif len(args) == 1:
            a = args[0]
        else:
            raise XPathSyntaxException('string() expects 1 argument or none')

        if isinstance(a, list): # node-set
            if len(a) == 0:
                return ''
            from ..Document import Document
            first_node = Document.ordered_first(a)
            return first_node.get_string_value()
        elif isinstance(a, Node) and hasattr(a, 'get_string_value'):
            return a.get_string_value()
        elif isinstance(a, float):
            if math.isnan(a):
                return 'NaN'
            elif a == -math.inf:
                return '-Infinity'
            elif a == math.inf:
                return 'Infinity'
            else:
                return str(a)
        elif a == True:
            return 'true'
        elif a == False:
            return 'false'
        else:
            return str(a)

    def f_concat(args, context_node, context_position, context_size, variables):
        if len(args) < 2:
            raise XPathSyntaxException('concat() expects at least 2 arguments')

        r = ''
        for a in args:
            r += a
        return r

    def f_starts_with(args, context_node, context_position, context_size, variables):
        if len(args) != 2:
            raise XPathSyntaxException('starts-with() expects 2 arguments')

        return args[0].startswith(args[1])

    def f_contains(args, context_node, context_position, context_size, variables):
        if len(args) != 2:
            raise XPathSyntaxException('contains() expects 2 arguments')

        return args[1] in args[0]

    def f_substring_before(args, context_node, context_position, context_size, variables):
        if len(args) != 2:
            raise XPathSyntaxException('substring-before() expects 2 arguments')

        return args[0].partition(args[1])[0]

    def f_substring_after(args, context_node, context_position, context_size, variables):
        if len(args) != 2:
            raise XPathSyntaxException('substring-after() expects 2 arguments')

        return args[0].partition(args[1])[2]

    def f_substring(args, context_node, context_position, context_size, variables):
        if len(args) not in (2,3):
            raise XPathSyntaxException('substring() expects 2 or 3 arguments')

        if args[1] == -math.inf:
            return ''

        arg_1 = Function.f_round((args[1],), context_node, context_position, context_size, variables)
        start = arg_1 - 1
        if start < 0:
            start = 0
        logger.debug('Substring start: ' + str(start))

        if len(args) > 2:
            if args[2] == -math.inf:
                return ''
            elif args[2] == math.inf:
                return args[0][start:]

            end = arg_1 - 1 + Function.f_round((args[2],), context_node, context_position, context_size, variables)
            logger.debug('Substring end: ' + str(end))

            return args[0][start:end]
        else:
            return args[0][start:]

    def f_string_length(args, context_node, context_position, context_size, variables):
        if len(args) != 1:
            raise XPathSyntaxException('string-length() expects 1 argument')
        if not isinstance(args[0], str):
            raise XPathSyntaxException('string-length() expects a string argument')

        return len(args[0])

    def f_normalize_space(args, context_node, context_position, context_size, variables):
        if len(args) == 0:
            a = context_node.get_string_value()
        elif len(args) == 1:
            a = args[0]
        else:
            raise XPathSyntaxException('normalize-space() expects 1 argument or none')

        a = a.strip('\x20\x09\x0D\x0A')
        return re.sub(r'[\x20\x09\x0D\x0A]+', ' ', a)

    def f_translate(args, context_node, context_position, context_size, variables):
        if len(args) != 3:
            raise XPathSyntaxException('translate() expects 3 arguments')

        s = args[0]
        from_ = args[1]
        to = args[2]
        if len(to) > len(from_):
            to = to[:len(from_)]
        delete = from_[len(to):]
        from_ = from_[:len(to)]

        return s.translate(str.maketrans(from_, to, delete))

    # Boolean Functions

    def f_boolean(args, context_node, context_position, context_size, variables):
        if len(args) != 1:
            raise XPathSyntaxException('boolean() expects 1 argument')

        if isinstance(args[0], int) or isinstance(args[0], float):
            if args[0] != 0 and not math.isnan(args[0]):
                return True
            else:
                return False
        elif isinstance(args[0], list):
            return len(args[0]) > 0
        elif isinstance(args[0], str):
            return len(args[0]) > 0
        else:
            return bool(args[0])

    def f_not(args, context_node, context_position, context_size, variables):
        if len(args) != 1:
            raise XPathSyntaxException('not() expects 1 argument')

        return not args[0]

    def f_true(args, context_node, context_position, context_size, variables):
        if len(args) != 0:
            raise XPathSyntaxException('true() accepts no arguments')

        return True

    def f_false(args, context_node, context_position, context_size, variables):
        if len(args) != 0:
            raise XPathSyntaxException('false() accepts no arguments')

        return False

    def f_lang(args, context_node, context_position, context_size, variables):
        if len(args) != 1:
            raise XPathSyntaxException('lang() expects 1 argument')

        a = args[0].lower()

        n = context_node
        while(not hasattr(n, 'attributes') or 'xml:lang' not in n.attributes):
            if n._parent is None:
                return False
            n = n._parent
        attr = n.attributes['xml:lang'].lower()

        if attr.startswith(a):
            return True
        else:
            return False

    # Number Functions

    def f_number(args, context_node, context_position, context_size, variables):
        if len(args) != 1:
            raise XPathSyntaxException('number() expects 1 argument')

        try:
            if isinstance(args[0], str):
                if args[0] == 'NaN':
                    return math.nan
                elif args[0] == 'Infinity':
                    return math.inf
                elif '.' in args[0]:
                    return float(args[0])
                else:
                    return int(args[0])
            elif args[0] == True:
                return 1
            elif args[0] == False:
                return 0
            elif isinstance(args[0],  list):
                s = f_string((args[0],), context_node, context_position, context_size, variables)
                return f_number(s, context_node, context_position, context_size, variables)
            elif isinstance(args[0], int) or isinstance(args[0], float):
                return args[0]
            else:
                return int(args[0])
        except ValueError:
            raise XPathSyntaxException('Invalid syntax for a number')

    def f_sum(args, context_node, context_position, context_size, variables):
        if len(args) != 1:
            raise XPathSyntaxException('sum() expects 1 argument')

        r = 0
        for n in args[0]:
            s = Function.f_string((n,), context_node, context_position, context_size, variables)
            i = Function.f_number((s,), context_node, context_position, context_size, variables)
            r += i
        return r

    def f_floor(args, context_node, context_position, context_size, variables):
        if len(args) != 1:
            raise XPathSyntaxException('floor() expects 1 argument')

        return math.floor(args[0])

    def f_ceiling(args, context_node, context_position, context_size, variables):
        if len(args) != 1:
            raise XPathSyntaxException('ceiling() expects 1 argument')

        return math.ceil(args[0])

    def f_round(args, context_node, context_position, context_size, variables):
        if len(args) != 1:
            raise XPathSyntaxException('round() expects 1 argument')

        if math.isnan(args[0]):
            return math.nan
        elif args[0] == math.inf:
            return math.inf
        elif args[0] == - math.inf:
            return - math.inf
        return round(args[0])

    FUNCTIONS = {
        # Node Set Functions
        'last': f_last,
        'position': f_position,
        'count': f_count,
        'id': f_id,
        'local-name': f_local_name,
        'namespace-uri': f_namespace_uri,
        'name': f_name,

        # String Functions
        'string': f_string,
        'concat': f_concat,
        'starts-with': f_starts_with,
        'contains': f_contains,
        'substring-before': f_substring_before,
        'substring-after': f_substring_after,
        'substring': f_substring,
        'string-length': f_string_length,
        'normalize-space': f_normalize_space,
        'translate': f_translate,

        # Boolean Functions
        'boolean': f_boolean,
        'not': f_not,
        'true': f_true, # x is to swallow the empty expression
        'false': f_false, # x is to swallow the empty expression
        'lang': f_lang,

        # Number Functions
        'number': f_number,
        'sum': f_sum,
        'floor': f_floor,
        'ceiling': f_ceiling,
        'round': f_round,
    }

    def __init__(self, name, function):
        self.name = name
        self.function = function
        self.children = []

    def evaluate(self, context_node, context_position, context_size, variables):
        arg_evals = []
        for c in self.children:
            v = c.evaluate(context_node, context_position, context_size, variables)
            logger.debug('Evaluated child of ' + str(self) + ' to ' + str(v))
            arg_evals.append(v)
        return self.function(arg_evals, context_node, context_position, context_size, variables)

    def __str__(self):
        return 'Function ' + self.name + ': [' + ','.join([str(x) for x in self.children]) + ']'



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import binascii
import logging
import re

from ..decorators import *
from .AnySimpleType import AnySimpleType

logger = logging.getLogger(__name__)

class HexBinaryType(AnySimpleType):
    def parse_value(self, value):
        value = super().parse_value(value)

        m = re.fullmatch(b'([0-9a-fA-F]{2})*', value)
        if not m:
            raise ValueError('xs:HexBinary must match ([0-9a-fA-F]{2})*')

        return binascii.a2b_hex(value)

    def produce_value(self, value):
        return binascii.b2a_hex(value)



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .NodeTest import NodeTest

logger = logging.getLogger(__name__)

class AnyNodeTest(NodeTest):
    def __init__(self, principal_node_type):
        super().__init__()
        self._prinicpal_node_type = principal_node_type

    def evaluate(self, context_node, context_position, context_size, variables):
        return context_node.get_type() == self._prinicpal_node_type

    def __str__(self):
        return 'AnyNodeTest ' + hex(id(self)) + ': [' + ','.join([str(x) for x in self.children]) + ']'



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging
import math

from .Function import Function
from ..exceptions import XPathSyntaxException

logger = logging.getLogger(__name__)
class Operator(object):
    def op_div(left, right):
        if right == 0:
            if left > 0:
                return math.inf
            else:
                return -math.inf
        else:
            return left // right

    OPERATORS = {
        '*': lambda x,y: x * y,
        # /
        # //
        # |
        '+': lambda x,y: x + y,
        '-': lambda x,y: x - y,
        '=': lambda x,y: x == y,
        '!=': lambda x,y: x != y,
        '<': lambda x,y: x < y,
        '<=': lambda x,y: x <= y,
        '>': lambda x,y: x > y,
        '>=': lambda x,y: x >= y,
        'and': lambda x,y: x and y,
        'or': lambda x,y: x or y,
        'mod': lambda x,y: math.fmod(x, y),
        'div': op_div,
        'negate': lambda x: - x,
    }
    def __init__(self, op):
        self.op = op
        self.children = []

    def evaluate(self, context_node, context_position, context_size, variables):
        from .Axis import Axis

        if self.op == 'negate':
            left = self.children[0].evaluate(context_node, context_position, context_size, variables)
            if isinstance(left, list):
                raise XPathSyntaxException('Got negate operator with a nodeset')
            elif isinstance(left, str):
                left = Functions.f_number(left)
            elif isinstance(left, bool):
                return not left
            elif isinstance(left, int) or isinstance(left, float):
                pass
            else:
                raise XPathSyntaxException('Unknown operand: ' + str(left))

            logger.debug('Negating ' + str(left))
            return Operator.OPERATORS['negate'](left)
        else:
            left = self.children[0].evaluate(context_node, context_position, context_size, variables)
            right = self.children[1].evaluate(context_node, context_position, context_size, variables)

            if self.op in ['or', 'and']:
                left = Function.f_boolean((left,), context_node, context_position, context_size, variables)
                right = Function.f_boolean((right,), context_node, context_position, context_size, variables)
            elif self.op in ['+', '-', 'div', 'mod']:
                left = Function.f_number((left,), context_node, context_position, context_size, variables)
                right = Function.f_number((right,), context_node, context_position, context_size, variables)

            if isinstance(left, list) and isinstance(right, list):
                for l in left:
                    l = Function.f_string((l,), context_node, context_position, context_size, variables)
                    for r in right:
                        r = Function.f_string((r,), context_node, context_position, context_size, variables)
                        logger.debug('Operator ' + str(l) + self.op + str(r))
                        if Operator.OPERATORS[self.op](l, r):
                            return True
                return False
            elif isinstance(left, list):
                if isinstance(right, int) or isinstance(right, float):
                    for x in left:
                        x = Function.f_number(Function.f_string((x,), context_node, context_position, context_size, variables))
                        logger.debug('Operator ' + str(x) + self.op + str(right))
                        if Operator.OPERATORS[self.op](x, right):
                            return True
                    return False
                elif isinstance(right, str):
                    for x in left:
                        x = Function.f_string((x,), context_node, context_position, context_size, variables)
                        logger.debug('Operator ' + str(x) + self.op + str(right))
                        if Operator.OPERATORS[self.op](x, right):
                            return True
                    return False
                elif isinstance(right, bool):
                    left = Function.f_boolean((left,), context_node, context_position, context_size, variables)
                    logger.debug('Operator ' + str(left) + self.op + str(right))
                    return Operator.OPERATORS[self.op](left, right)
                else:
                    raise XPathSyntaxException('Unknown right hand operand: ' + str(right))
            elif isinstance(right, list):
                if isinstance(left, int) or isinstance(left, float):
                    for x in right:
                        x = Function.f_number(Function.f_string((x,), context_node, context_position, context_size, variables))
                        logger.debug('Operator ' + str(left) + self.op + str(x))
                        if Operator.OPERATORS[self.op](left, x):
                            return True
                    return False
                elif isinstance(left, str):
                    for x in right:
                        x = Function.f_string((x,), context_node, context_position, context_size, variables)
                        logger.debug('Operator ' + str(left) + self.op + str(x))
                        if Operator.OPERATORS[self.op](left, x):
                            return True
                    return False
                elif isinstance(left, bool):
                    right = Function.f_boolean((right,), context_node, context_position, context_size, variables)
                    logger.debug('Operator ' + str(left) + self.op + str(right))
                    return Operator.OPERATORS[self.op](left, right)
                else:
                    raise XPathSyntaxException('Unknown left hand operand: ' + str(left))
            else:
                if isinstance(left, bool) or isinstance(right, bool):
                    left = Function.f_boolean((left,), context_node, context_position, context_size, variables)
                    right = Function.f_boolean((right,), context_node, context_position, context_size, variables)
                    logger.debug('Operator ' + str(left) + self.op + str(right))
                    return Operator.OPERATORS[self.op](left, right)
                elif isinstance(left, int) or isinstance(left, float) \
                or isinstance(right, int) or isinstance(right, float):
                    left = Function.f_number((left,), context_node, context_position, context_size, variables)
                    right = Function.f_number((right,), context_node, context_position, context_size, variables)
                    logger.debug('Operator ' + str(left) + self.op + str(right))
                    return Operator.OPERATORS[self.op](left, right)
                elif isinstance(left, str) or isinstance(right, str):
                    left = Function.f_string((left,), context_node, context_position, context_size, variables)
                    right = Function.f_string((right,), context_node, context_position, context_size, variables)
                    logger.debug('Operator ' + str(left) + self.op + str(right))
                    return Operator.OPERATORS[self.op](left, right)
                else:
                    return Operator.OPERATORS[self.op](left, right)

    def __str__(self):
        return 'Operator ' + self.op + ' ' + hex(id(self)) + ': [' + ','.join([str(x) for x in self.children]) + ']'



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

from ..exceptions import XPathSyntaxException
from .AnyNodeTest import AnyNodeTest
from .Axis import Axis
from .Expression import Expression
from .Function import Function
from .Literal import Literal
from .NCNameNodeTest import NCNameNodeTest
from .NodeTest import NodeTest
from .Operator import Operator
from .Predicate import Predicate
from .QNameNodeTest import QNameNodeTest
from .RootStep import RootStep
from .Step import Step
from .TypeNodeTest import TypeNodeTest

__all__ = [
    'AnyNodeTest',
    'Axis',
    'Expression',
    'Function',
    'Literal',
    'NCNameNodeTest',
    'NodeTest',
    'Operator',
    'Predicate',
    'QNameNodeTest',
    'RootStep',
    'Step',
    'TypeNodeTest',

    'XPathSyntaxException',
]



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .exceptions import *
from .Node import Node

logger = logging.getLogger(__name__)

class Parent(Node):
    '''
    Super class for nodes containing children

    :param parent: parent Node of this Node
    :type parent: Parent or None

    '''
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.children = []

    def spawn_character_data(self, *args, **kwargs):
        '''
        Spawn a :py:class:`.CharacterData` object using this node as the parent

        All arguments are passed to the newly created object's constructor
        '''
        from .CharacterData import CharacterData
        n = CharacterData(*args, **kwargs, parent=self)

        self.children.append(n)

        return n

    def spawn_comment(self, *args, **kwargs):
        '''
        Spawn a :py:class:`.Comment` object using this node as the parent

        All arguments are passed to the newly created object's constructor
        '''
        from .Comment import Comment
        n = Comment(*args, **kwargs, parent=self)

        self.children.append(n)

        return n

    def spawn_element(self, *args, **kwargs):
        '''
        Spawn a :py:class:`.Element` object using this node as the parent

        All arguments are passed to the newly created object's constructor
        '''
        from .Element import Element
        n = Element(*args, **kwargs, parent=self)

        self.children.append(n)

        return n

    def spawn_processing_instruction(self, *args, **kwargs):
        '''
        Spawn a :py:class:`.ProcessingInstruction` object using this node as the parent

        All arguments are passed to the newly created object's constructor
        '''
        from .ProcessingInstruction import ProcessingInstruction
        n = ProcessingInstruction(*args, **kwargs, parent=self)

        self.children.append(n)

        return n

    def __len__(self):
        return len(self.children)

    def __getitem__(self, key):
        if not isinstance(key, int) and not isinstance(key, slice):
            raise TypeError('Key values must be of int type or slice; got: ' + key.__class__.__name__)

        return self.children[key]

    def __setitem__(self, key, value):
        if not isinstance(key, int):
            raise TypeError('Key values must be of int type; got: ' + key.__class__.__name__)
        if not isinstance(value, Node):
            raise TypeError('Values must be of Node type; got: ' + value.__class__.__name__)

        self.children[key] = value

    def __delitem__(self, key):
        if not isinstance(key, int):
            raise TypeError('Key values must be of int type; got: ' + key.__class__.__name__)

        del self.children[key]

    def __iter__(self):
        return iter(self.children)

    def append(self, x):
        from .CharacterData import CharacterData

        if isinstance(x, str):
            # wrap in CharaterData
            n = CharacterData(x, parent=self)
        elif isinstance(x, int) or isinstance(x, float):
            # convert to str & wrap in CharaterData
            n = CharacterData(str(x), parent=self)
        elif isinstance(x, Node):
            n = x
            n._parent = self
        else:
            raise ValueError('Children of ' + self.__class__.__name__
                + ' must be a simple type (str, int, float)'
                + ' or a subclass of Node; got: ' + x.__class__.__name__)

        self.children.append(n)

    def count(self):
        return self.children.count()

    def index(self, x):
        return self.children.index(x)

    def extend(self, iterable):
        for c in iterable:
            self.append(c)

    def insert(self, i, x):
        from .CharacterData import CharacterData

        if isinstance(x, str):
            # wrap in CharaterData
            n = CharacterData(x, parent=self)
        elif isinstance(x, int) or isinstance(x, float):
            # convert to str & wrap in CharaterData
            n = CharacterData(str(x), parent=self)
        elif isinstance(x, Node):
            n = x
        else:
            raise ValueError('Children of ' + self.__class__.__name__ + ' must be subclass of Node; got: ' + x.__class__.__name__)

        self.children.insert(i, n)

    def pop(self, i=-1):
        n = self.children.pop(i)
        self.detach(n)
        return n

    def remove(self, x):
        n = self.children[self.children.index(x)]
        self.children.remove(x)

    def reverse(self):
        return self.children.reverse()

    def sort(self, key=None, reverse=False):
        return self.children.sort(key=key, reverse=reverse)

    # TODO copy()

    def find_by_id(self, id_):
        logger.debug(str(self) + ' checking children for id: ' + str(id_))
        for c in self.children:
            el = c.find_by_id(id_)
            if el is not None:
                return el

        return super().find_by_id(id_)



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from ..decorators import *
from .AnnotatedType import AnnotatedType
from .ChoiceElement import ChoiceElement
from .FacetType import FacetType
from .GroupType import GroupType
from .PatternElement import PatternElement
from .QNameType import QNameType
from .TotalDigitsElement import TotalDigitsElement
from .WhitespaceElement import WhitespaceElement
from .WildcardType import WildcardType

logger = logging.getLogger(__name__)

@attribute(local_name='base', type=QNameType)
@element(local_name='group', list='tags', cls=GroupType, min=0)
@element(local_name='all', list='tags',
    cls=('expatriate.model.xs.AllType', 'AllType'),
    min=0)
@element(local_name='choice', list='tags', cls=ChoiceElement, min=0)
@element(local_name='sequence', list='tags', cls=GroupType, min=0)
@element(local_name='simpleType', list='tags',
    cls=('expatriate.model.xs.SimpleTypeType', 'SimpleTypeType'),
    min=0)
@element(local_name='minExclusive', list='tags', cls=FacetType, min=0, max=None)
@element(local_name='minInclusive', list='tags', cls=FacetType, min=0, max=None)
@element(local_name='maxExclusive', list='tags', cls=FacetType, min=0, max=None)
@element(local_name='maxInclusive', list='tags', cls=FacetType, min=0, max=None)
@element(local_name='totalDigits', list='tags', cls=TotalDigitsElement, min=0, max=None)
@element(local_name='fractionDigits', list='tags', cls=FacetType, min=0, max=None)
@element(local_name='length', list='tags', cls=FacetType, min=0, max=None)
@element(local_name='minLength', list='tags', cls=FacetType, min=0, max=None)
@element(local_name='maxLength', list='tags', cls=FacetType, min=0, max=None)
@element(local_name='enumeration', list='tags', cls=FacetType, min=0, max=None)
@element(local_name='whiteSpace', list='tags', cls=WhitespaceElement, min=0, max=None)
@element(local_name='pattern', list='tags', cls=PatternElement, min=0, max=None)
@element(local_name='attribute', list='tags',
    cls=('expatriate.model.xs.AttributeType', 'AttributeType'),
    min=0, max=None)
@element(local_name='attributeGroup', list='tags',
    cls=('expatriate.model.xs.AttributeGroupType', 'AttributeGroupType'),
    min=0, max=None)
@element(local_name='anyAttribute', list='tags', cls=WildcardType, min=0)
class RestrictionType(AnnotatedType):
    def get_defs(self, schema, top_level):
        logger.debug('Base: ' + self.base)
        # TODO unable to map xmlns because ET doesn't retain it
        base_ns, base_name = [self.base.partition(':')[i] for i in [0,2]]
        top_level.set_super_module(base_ns)
        top_level.set_super_class(base_name)

        return super().get_defs(schema, top_level)



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

from .xs.AllNniType import AllNniType
from .xs.AnySimpleType import AnySimpleType
from .xs.AnyTypeType import AnyTypeType
from .xs.AnyUriType import AnyUriType
from .xs.Base64BinaryType import Base64BinaryType
from .xs.BooleanType import BooleanType
from .xs.ByteType import ByteType
from .xs.DateTimeStampType import DateTimeStampType
from .xs.DateTimeType import DateTimeType
from .xs.DateType import DateType
from .xs.DayTimeDurationType import DayTimeDurationType
from .xs.DecimalType import DecimalType
from .xs.DoubleType import DoubleType
from .xs.DurationType import DurationType
from .xs.EntitiesType import EntitiesType
from .xs.EntityType import EntityType
from .xs.FloatType import FloatType
from .xs.GDayType import GDayType
from .xs.GMonthDayType import GMonthDayType
from .xs.GMonthType import GMonthType
from .xs.GYearMonthType import GYearMonthType
from .xs.GYearType import GYearType
from .xs.HexBinaryType import HexBinaryType
from .xs.IdRefsType import IdRefsType
from .xs.IdRefType import IdRefType
from .xs.IdType import IdType
from .xs.IntegerType import IntegerType
from .xs.IntType import IntType
from .xs.LanguageType import LanguageType
from .xs.LongType import LongType
from .xs.NamespaceListType import NamespaceListType
from .xs.NameType import NameType
from .xs.NCNameType import NCNameType
from .xs.NegativeIntegerType import NegativeIntegerType
from .xs.NMTokensType import NMTokensType
from .xs.NMTokenType import NMTokenType
from .xs.NonNegativeIntegerType import NonNegativeIntegerType
from .xs.NonPositiveIntegerType import NonPositiveIntegerType
from .xs.NormalizedStringType import NormalizedStringType
from .xs.PositiveIntegerType import PositiveIntegerType
from .xs.QNameType import QNameType
from .xs.ShortType import ShortType
from .xs.StringType import StringType
from .xs.TimeType import TimeType
from .xs.TokenType import TokenType
from .xs.UnsignedByteType import UnsignedByteType
from .xs.UnsignedIntType import UnsignedIntType
from .xs.UnsignedLongType import UnsignedLongType
from .xs.UnsignedShortType import UnsignedShortType
from .xs.YearMonthDurationType import YearMonthDurationType



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .Node import Node

logger = logging.getLogger(__name__)

class CharacterData(Node):
    '''
    Class representing XML character data
    '''
    def __init__(self, data, cdata_block=False, parent=None):
        super().__init__(parent=parent)

        self.data = data
        self.cdata_block = cdata_block

    def produce(self):
        s = ''
        if self.cdata_block:
            s += '<![CDATA[' + self.data.replace(']]>', ']]&gt;') + ']]>'
        else:
            s += self.escape(self.data)
        return s

    def get_string_value(self):
        '''
        Return the string value of the node
        '''
        return self.data

    def get_type(self):
        '''
        Return the type of the node
        '''
        return 'text'

    def __eq__(self, other):
        if isinstance(other, str):
            return self.data == other
        elif isinstance(other, int) or isinstance(other, float):
            return self.data == str(other)
        else:
            return object.__eq__(self, other)



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re

from ..decorators import *
from .DurationType import DurationType

logger = logging.getLogger(__name__)

class DayTimeDurationType(DurationType):
    def parse_value(self, value):
        m = re.fullmatch(r'-?P(\d+D)?(T(\d+H)?(\d+M)?(\d+(\.\d+)?S)?)?', value)
        if not m or not re.fullmatch(r'.*[DHMS].*', value) or not re.fullmatch(r'.*[^T]', value):
            raise ValueError('Unable to parse xs:DayTimeDurationType value')

        return super().parse_value(value)

    def produce_value(self, value):
        months, seconds = value
        if months != 0:
            raise ValueError('xs:DayTimeDurationType requires 0 for months value')

        return super().produce_value(value)



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re

from ..decorators import *
from .AnySimpleType import AnySimpleType

logger = logging.getLogger(__name__)

class DurationType(AnySimpleType):
    def parse_value(self, value):
        m = re.fullmatch(r'-?P(\d+Y)?(\d+M)?(\d+D)?(T(\d+H)?(\d+M)?(\d+(\.\d+)?S)?)?', value)
        if not m or not re.fullmatch(r'.*[YMDHS].*', value) or not re.fullmatch(r'.*[^T]', value):
            raise ValueError('Unable to parse xs:Duration value')

        if value.startswith('-'):
            signed = True
        else:
            signed = False

        # fudge for python 3.5 compatibility w/match obj __getitem__
        m = list(m.groups())
        m.insert(0, None)

        if m[1] is not None:
            years = int(m[1].strip('Y'))
        else:
            years = 0

        if m[2] is not None:
            months = int(m[2].strip('M'))
        else:
            months = 0

        if m[3] is not None:
            days = int(m[3].strip('D'))
        else:
            days = 0

        if m[5] is not None:
            hours = int(m[5].strip('H'))
        else:
            hours = 0

        if m[6] is not None:
            minutes = int(m[6].strip('M'))
        else:
            minutes = 0

        if m[7] is not None:
            seconds = float(m[7].strip('S'))
        else:
            seconds = 0.0

        # collapse years into months
        months += years * 12

        # collapse the rest into seconds
        hours += days * 24
        minutes += hours * 60
        seconds += minutes * 60.0

        if signed:
            return (- months, - seconds)
        else:
            return (months, seconds)

    def produce_value(self, value):
        if not isinstance(value, tuple):
            raise ValueError('xs:duration is produced from a months, seconds tuple')

        months, seconds = value

        if months < 0 or seconds < 0:
            r = '-P'
            months = - months
            seconds = - seconds
        else:
            r = 'P'

        years, months = divmod(months, 12)
        minutes, seconds = divmod(seconds, 60)
        hours, minutes = divmod(int(minutes), 60)
        days, hours = divmod(hours, 24)

        if years != 0:
            r += '%dY' % years
        if months != 0:
            r += '%dM' % months
        if days != 0:
            r += '%dD' % days
        if hours != 0 or minutes != 0 or seconds != 0.0:
            r += 'T'
            if hours != 0:
                r += '%dH' % hours
            if minutes != 0:
                r += '%dM' % minutes
            if seconds != 0.0:
                # be nice if we could print based on float's precision,
                # but we just cut off if it's an int value
                seconds_float = seconds - int(seconds)
                if seconds_float != 0.0:
                    r += '%fS' % seconds
                else:
                    r += '%dS' % int(seconds)

        return r



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .NodeTest import NodeTest

logger = logging.getLogger(__name__)
class TypeNodeTest(NodeTest):
    NODE_TYPES = [
        'comment',
        'text',
        'processing-instruction',
        'node',
    ]

    def __init__(self, name):
        super().__init__()
        self.name = name

    # TODO processing-instruction can have a literal that matches its name

    def evaluate(self, context_node, context_position, context_size, variables):
        if self.name == 'node':
            return True
        else:
            return context_node.get_type() == self.name

    def __str__(self):
        return 'TypeNodeTest ' + hex(id(self)) + ' ' + self.name + ': [' + ','.join([str(x) for x in self.children]) + ']'



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .Step import Step

logger = logging.getLogger(__name__)
class RootStep(Step):
    def __init__(self, document):
        self.children = []
        self._document = document

    def evaluate(self, context_node, context_position, context_size, variables):
        if len(self.children) == 0:
            logger.debug('Root step with no children: using ' + str(self._document) + ' as the result set')
            return [self._document]
        else:
            return super().evaluate(self._document, 1, 1, variables)

    def __str__(self):
        return 'RootStep ' + hex(id(self)) + ': [' + ','.join([str(x) for x in self.children]) + ']'



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .NodeTest import NodeTest
from .Predicate import Predicate

logger = logging.getLogger(__name__)
class Axis(object):
    def a_ancestor(node):
        ns = []
        while(node._parent is not None):
            ns.append(node._parent)
            node = node._parent
        return ns

    def a_ancestor_or_self(node):
        ns = [node]
        ns.extend(Axis.a_ancestor(node))
        return ns

    def a_attribute(node):
        if hasattr(node, 'attribute_nodes'):
            return list(node.attribute_nodes.values())
        else:
            return []

    def a_child(node):
        if hasattr(node, 'children'):
            return node.children.copy()
        else:
            return []

    def a_descendant(node):
        logger.debug('Collecting descendants of ' + str(node))
        if not hasattr(node, 'children') or len(node.children) == 0:
            logger.debug(str(node) + ' has no children')
            return []
        else:
            ns = []
            for n in node.children:
                ns.append(n)
                ns.extend(Axis.a_descendant(n))
            logger.debug(str(node) + ' descendants: [' + ','.join([str(x) for x in ns]) + ']')
            return ns

    def a_descendant_or_self(node):
        ns = [node]
        ns.extend(Axis.a_descendant(node))
        return ns

    def a_namespace(node):
        if hasattr(node, 'namespace_nodes'):
            return list(node.namespace_nodes.values())
        else:
            return []

    def a_parent(node):
        if node._parent is not None:
            return [node._parent]
        else:
            return []

    def a_following_sibling(node):
        from ..Attribute import Attribute
        from ..Namespace import Namespace
        if isinstance(node, Attribute) \
        or isinstance(node, Namespace) \
        or node._parent is None:
            return []

        ns = []
        # figure out our index then go through children > our index
        node_i = node._parent.children.index(node) + 1
        for n in node._parent.children[node_i:]:
            ns.append(n)
        return ns

    def a_following(node):
        from ..Attribute import Attribute
        from ..Namespace import Namespace
        if isinstance(node, Attribute) \
        or isinstance(node, Namespace):
            return []

        ns = []
        while node is not None:
            ns.extend(Axis.a_following_sibling(node))
            node = node._parent
        return ns

    def a_preceding_sibling(node):
        from ..Attribute import Attribute
        from ..Namespace import Namespace
        if isinstance(node, Attribute) \
        or isinstance(node, Namespace) \
        or node._parent is None:
            return []

        ns = []
        # figure out our index then go through children < our index
        node_i = node._parent.children.index(node)
        for n in reversed(node._parent.children[:node_i]):
            ns.append(n)
        return ns

    def a_preceding(node):
        from ..Attribute import Attribute
        from ..Namespace import Namespace
        if isinstance(node, Attribute) \
        or isinstance(node, Namespace):
            return []

        ns = []
        while node is not None:
            logger.debug('Extending nodeset with ' + str(node) + ' preceding siblings')
            ns.extend(Axis.a_preceding_sibling(node))
            if node._parent is not None:
                logger.debug('Appending parent of ' + str(node) + ': ' + str(node._parent))
                ns.append(node._parent)
            node = node._parent
        return ns

    AXES = {
        'ancestor': a_ancestor,
        'ancestor-or-self': a_ancestor_or_self,
        'attribute': a_attribute,
        'child': a_child,
        'descendant': a_descendant,
        'descendant-or-self': a_descendant_or_self,
        'following': a_following,
        'following-sibling': a_following_sibling,
        'namespace': a_namespace,
        'parent': a_parent,
        'preceding': a_preceding,
        'preceding-sibling': a_preceding_sibling,
        'self': lambda node: [node],
    }

    PRINCIPAL_NODE_TYPE = {
        'ancestor': 'element',
        'ancestor-or-self': 'element',
        'attribute': 'attribute',
        'child': 'element',
        'descendant': 'element',
        'descendant-or-self': 'element',
        'following': 'element',
        'following-sibling': 'element',
        'namespace': 'namespace',
        'parent': 'element',
        'preceding': 'element',
        'preceding-sibling': 'element',
        'self': 'element',
    }

    AXIS_DIRECTION = {
        'ancestor': 'reverse',
        'ancestor-or-self': 'reverse',
        'attribute': 'forward',
        'child': 'forward',
        'descendant': 'forward',
        'descendant-or-self': 'forward',
        'following': 'forward',
        'following-sibling': 'forward',
        'namespace': 'forward',
        'parent': 'forward',
        'preceding': 'reverse',
        'preceding-sibling': 'reverse',
        'self': 'forward',
    }

    def __init__(self, name):
        self.name = name
        self.children = []

    def evaluate(self, context_node, context_position, context_size, variables):
        from ..Document import Document
        if len(self.children) <= 0:
            raise ValueError('Axis missing NodeTest')
        for i, child in enumerate(self.children):
            if i == 0 and not isinstance(child, NodeTest):
                raise ValueError('Axis missing NodeTest')
            elif i > 0 and not isinstance(child, Predicate):
                raise ValueError('Axis children past the first must be predicates: ' + str(child))

        nodeset = Axis.AXES[self.name](context_node)
        if len(nodeset) == 0:
            return nodeset

        logger.debug('Initial nodeset: [' + ','.join([str(x) for x in nodeset]) + ']')
        for c in self.children:
            ns = []
            for i, n in enumerate(nodeset):
                # logger.debug('Testing ' + str(nodeset[i]) + ' against ' + str(c))
                if c.evaluate(n, i+1, len(nodeset), variables):
                    # logger.debug(str(nodeset[i]) + ' passed ' + str(c))
                    ns.append(n)
                # else:
                #     logger.debug(str(nodeset[i]) + ' failed ' + str(c))
            nodeset = ns
        if Axis.AXIS_DIRECTION[self.name] == 'forward':
            nodeset = Document.order_sort(nodeset)
        else:
            nodeset = Document.order_sort(nodeset, reverse=True)
        logger.debug('Final nodeset: [' + ','.join([str(x) for x in nodeset]) + ']')

        return nodeset

    def get_principal_node_type(self):
        return Axis.PRINCIPAL_NODE_TYPE[self.name]

    def __str__(self):
        return 'Axis ' + hex(id(self)) + ' ' + self.name + ': [' + ','.join([str(x) for x in self.children]) + ']'



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import logging

logger = logging.getLogger(__name__)

class Mapper(object):
    '''
    Super class of Mapper types. Defines the methods to be overriden and a few
    convenience methods.
    '''
    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def __str__(self):
        return self.__class__.__name__ + ' ' + str(self._kwargs)

    def _load_cls(self, cls_):
        '''
        Wrapper function to either return cls_ if it is a class or load the
        class if it is a tuple of the form (package, class_name)
        '''
        if isinstance(cls_, tuple):
            mod = importlib.import_module(cls_[0])
            return getattr(mod, cls_[1])
        else:
            return cls_

    def initialize(self, *args, **kwargs):
        raise NotImplementedError

    def matches(self, *args, **kwargs):
        raise NotImplementedError

    def parse_in(self, *args, **kwargs):
        raise NotImplementedError

    def validate(self, *args, **kwargs):
        raise NotImplementedError

    def produce_in(self, *args, **kwargs):
        raise NotImplementedError



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re
import xml.parsers.expat

from .CharacterData import CharacterData
from .Comment import Comment
from .Element import Element
from .exceptions import *
from .Node import Node
from .Parent import Parent
from .ProcessingInstruction import ProcessingInstruction

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class Document(Parent):
    '''
    Class representing a XML document
    '''
    @staticmethod
    def ordered_first(nodeset):
        if len(nodeset) == 0:
            raise ValueError('Cannot determine the first node of an empty nodeset')
        first = nodeset[0]
        for n in nodeset[1:]:
            if n.get_document_order() < first.get_document_order():
                first = n
        return first

    @staticmethod
    def is_nodeset(nodeset):
        if not isinstance(nodeset, list):
            return False
        for n in nodeset:
            if not isinstance(n, Node):
                return False
        return True

    @staticmethod
    def order_sort(nodeset, reverse=False):
        if not Document.is_nodeset(nodeset):
            raise TypeError('Cannot sort by document order without a nodeset')
        for n in nodeset:
            logger.info(str(n) + ' document order: ' + str(n.get_document_order()))
        return sorted(nodeset, key=lambda n: n.get_document_order(), reverse=reverse)

    def __init__(self, encoding=None, skip_whitespace=True):
        super().__init__()
        self.version = None
        self.encoding = encoding
        self.standalone = None

        self.root_element = None

        self._parser = xml.parsers.expat.ParserCreate(encoding=encoding)
        self._skip_whitespace = skip_whitespace
        self._in_space_preserve = False
        self._in_cdata = False
        self._stack = []

        self._parser.XmlDeclHandler = self._xml_decl_handler
        self._parser.StartElementHandler = self._start_element_handler
        self._parser.EndElementHandler = self._end_element_handler
        self._parser.ProcessingInstructionHandler = self._processing_instruction_handler
        self._parser.CharacterDataHandler = self._character_data_handler
        self._parser.CommentHandler = self._comment_handler
        self._parser.StartCdataSectionHandler = self._start_cdata_section_handler
        self._parser.EndCdataSectionHandler = self._end_cdata_section_handler
        self._parser.DefaultHandlerExpand = self._default_handler_expand
        self._parser.NotStandaloneHandler = self._not_standalone_handler

    def parse(self, data, isfinal=True):
        logger.debug('Parsing data: ' + str(data))
        self._parser.Parse(data, isfinal)

        # TODO check that we're the only thing left on the stack when isfinal

    def parse_file(self, file_):
        logger.debug('Parsing file: ' + str(file_))
        self._parser.ParseFile(file_)

        # TODO check that we're the only thing left on the stack when isfinal

    def produce(self, xml_decl=True):
        if self.encoding is None:
            self.encoding = 'UTF-8'

        s = ''
        if xml_decl:
            s = '<?xml version="'
            if self.version is None:
                self.version = 1.0
            s += str(self.version)
            s += '"'
            s += ' encoding="' + self.encoding + '"'
            if self.standalone is not None:
                if self.standalone:
                    s += ' standalone="yes"'
                else:
                    s += ' standalone="no"'
            s += '>'

        for item in self.children:
            s += item.produce()

        return s.encode(self.encoding)

    def _xml_decl_handler(self, version, encoding, standalone):
        logger.debug('_xml_decl_handler version: ' + str(version) + ' encoding: ' + str(encoding) + ' standalone: ' + str(standalone))
        self.version = float(version)
        self.encoding = encoding
        if standalone is None or standalone == -1:
            self.standalone = None
        else:
            if standalone.lower() == 'yes':
                self.standalone = True
            else:
                self.standalone = False

    def _start_element_handler(self, name, attributes):
        logger.debug('_start_element_handler elname: ' + str(name) + ' attname: ' + str(name) + ' attributes: ' + str(attributes))

        # check for whitespace preservation
        if 'xml:space' in attributes and attributes['xml:space'] == 'preserve':
            self._in_space_preserve = True

        if ':' in name:
            prefix, colon, local_name = name.partition(':')
        else:
            prefix = None
            local_name = name

        if len(self._stack) == 0:
            el = Element(local_name, attributes, parent=self, prefix=prefix)
            self.root_element = el
            self.children.append(el)
        else:
            el = Element(local_name, attributes, parent=self._stack[-1], prefix=prefix)
            self._stack[-1].children.append(el)

        self._stack.append(el)

    def _end_element_handler(self, name):
        logger.debug('_end_element_handler name: ' + str(name))
        el = self._stack.pop()
        if el.name != name:
            raise ValueError('Stack pop element name (' + el.name + ') does not match end tag name: ' + name)

        # check for whitespace preservation
        if 'xml:space' in el.attributes and el.attributes['xml:space'] == 'preserve':
            self._in_space_preserve = False

    def _processing_instruction_handler(self, target, data):
        logger.debug('_processing_instruction_handler target: ' + str(target) + ' data: ' + str(data))

        if len(self._stack) == 0:
            pi = ProcessingInstruction(target, data, parent=self)
            self.children.append(pi)
        else:
            pi = ProcessingInstruction(target, data, parent=self._stack[-1])
            self._stack[-1].children.append(pi)

    def _character_data_handler(self, data):
        logger.debug('_character_data_handler data: ' + str(data.encode('UTF-8')))
        if not self._in_space_preserve:
            if self._skip_whitespace:
                data = data.strip(' \t\n')
                logger.debug('Stripped to: ' + str(data.encode('UTF-8')))
            if data == '':
                logger.debug('Skipping whitespace character data')
                return

        if len(self._stack) == 0:
            char_data = CharacterData(data, cdata_block=self._in_cdata, parent=self)
            self.children.append(char_data)
        else:
            char_data = CharacterData(data, cdata_block=self._in_cdata, parent=self._stack[-1])
            self._stack[-1].children.append(char_data)

    def _comment_handler(self, data):
        logger.debug('_comment_handler data: ' + str(data))

        if len(self._stack) == 0:
            c = Comment(data, parent=self)
            self.children.append(c)
        else:
            c = Comment(data, parent=self._stack[-1])
            self._stack[-1].children.append(c)

    def _start_cdata_section_handler(self):
        logger.debug('_start_cdata_section_handler')
        self._in_cdata = True

    def _end_cdata_section_handler(self):
        logger.debug('_end_cdata_section_handler')
        self._in_cdata = False

    def _default_handler_expand(self, data):
        logger.debug('_default_handler_expand data: ' + str(data))

    def _not_standalone_handler(self, data):
        logger.debug('_not_standalone_handler data: ' + str(data))

    def get_type(self):
        '''
        Return the type of the node
        '''
        return 'root'

    def get_string_value(self):
        '''
        Return the string value of the node
        '''
        return self.root_element.get_string_value()

    def get_document_order(self):
        return 0



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os

from ..decorators import *
from .AnnotationElement import AnnotationElement
from .AnyTypeType import AnyTypeType
from .IdType import IdType

logger = logging.getLogger(__name__)

@attribute(local_name='id', type=IdType)
@element(local_name='annotation', list='tags', cls=AnnotationElement, min=0)
class AnnotatedType(AnyTypeType):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._super_module = 'expatriate.model.Model'
        self._super_class = 'Model'
        self._value_enumeration = None
        self._value_pattern = None

    def set_super_module(self, super_module):
        self._super_module = super_module

    def set_super_class(self, super_class):
        self._super_class = super_class

    def append_value_enumeration(self, value):
        if self._value_enumeration is None:
            self._value_enumeration = []
        self._value_enumeration.append(value)

    def set_value_pattern(self, pattern):
        self._value_pattern = pattern

    def stub(self, path, schema, class_name):
        model_map = {'elements': [], 'attributes': {}}
        for t in self.tags:
            defs = t.get_defs(schema, self)
            model_map['elements'].extend(defs['elements'])
            model_map['attributes'].update(defs['attributes'])

        logger.debug('Stubbing ' + str(self) + ' to ' + class_name + '.py')

        with open(os.path.join(path, class_name + '.py'), 'w') as f:
            f.write(STUB_HEADER)
            f.write('import logging\n\n')
            f.write('from ' + self._super_module + ' import ' + self._super_class + '\n\n')
            f.write('logger = logging.getLogger(__name__)\n')
            f.write('class ' + class_name + '(' + self._super_class + '):\n')

            indent = 4

            f.write((' ' * indent) + "MODEL_MAP = {\n")
            indent += 4

            f.write((' ' * indent) + "'elements': [\n")
            indent += 4
            for el_def in model_map['elements']:
                f.write((' ' * indent) + str(el_def) + ",\n")
            indent -= 4
            f.write((' ' * indent) + "],\n")

            f.write((' ' * indent) + "'attributes': {\n")
            indent += 4
            for name, att_def in model_map['attributes']:
                f.write((' ' * indent) + "'" + name + "': " + str(att_def) + ",\n")
            indent -= 4
            f.write((' ' * indent) + "},\n")

            indent -= 4
            f.write((' ' * indent) + "}\n\n")

            if self._value_enumeration is not None:
                f.write((' ' * indent) + "def get_value_enum(self):\n")
                indent += 4
                f.write((' ' * indent) + "return [\n")
                indent += 4
                for v in self._value_enumeration:
                    f.write((' ' * indent) + repr(v) + ",\n")
                indent -= 4
                f.write((' ' * indent) + "]\n\n")
                indent -= 4

            if self._value_pattern is not None:
                f.write((' ' * indent) + "def get_value_pattern(self):\n")
                indent += 4
                f.write((' ' * indent) + "return r'" + self._value_pattern + "'\n\n")
                indent -= 4



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from ..decorators import *
from .AnnotatedType import AnnotatedType
from .AttributeGroupType import AttributeGroupType
from .AttributeType import AttributeType
from .ChoiceElement import ChoiceElement
from .GroupType import GroupType
from .QNameType import QNameType
from .WildcardType import WildcardType

logger = logging.getLogger(__name__)

@attribute(local_name='base', type=QNameType, required=True)
@element(local_name='group', list='tags', cls=GroupType, min=0)
@element(local_name='all', list='tags',
    cls=('expatriate.model.xs.AllType', 'AllType'),
    min=0)
@element(local_name='choice', list='tags', cls=ChoiceElement, min=0)
@element(local_name='sequence', list='tags', cls=GroupType, min=0)
@element(local_name='attribute', list='tags', cls=AttributeType, min=0, max=None)
@element(local_name='attributeGroup', list='tags', cls=AttributeGroupType, min=0, max=None)
@element(local_name='anyAttribute', list='tags', cls=WildcardType, min=0)
class ExtensionType(AnnotatedType):
    def get_defs(self, schema, top_level):
        logger.debug('Base: ' + self.base)
        # TODO unable to map xmlns because ET doesn't retain it
        base_ns, base_name = [self.base.partition(':')[i] for i in [0,2]]
        top_level.set_super_module(base_ns)
        top_level.set_super_class(base_name)

        return super().get_defs(schema, top_level)



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from ..decorators import *
from .AnnotatedType import AnnotatedType
from .AnySimpleType import AnySimpleType
from .BooleanType import BooleanType

logger = logging.getLogger(__name__)

@attribute(local_name='value', required=True, type=AnySimpleType)
@attribute(local_name='fixed', type=BooleanType, default=False)
@attribute(local_name='*', )
class FacetType(AnnotatedType):
    def get_defs(self, schema, top_level):
        if self.tag_name == 'enumeration':
            top_level.append_value_enumeration(self.value)

        return super().get_defs(schema, top_level)



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os.path

from . import FORM_CHOICE_ENUMERATION
from ..decorators import *
from .AnnotationElement import AnnotationElement
from .AnyTypeType import AnyTypeType
from .AnyUriType import AnyUriType
from .AttributeGroupType import AttributeGroupType
from .AttributeType import AttributeType
from .ComplexTypeType import ComplexTypeType
from .ElementType import ElementType
from .GroupType import GroupType
from .IdType import IdType
from .ImportElement import ImportElement
from .IncludeElement import IncludeElement
from .NotationElement import NotationElement
from .RedefineElement import RedefineElement
from .SimpleTypeType import SimpleTypeType
from .TokenType import TokenType

logger = logging.getLogger(__name__)

@attribute(local_name='targetNamespace', type=AnyUriType)
@attribute(local_name='version', type=TokenType)
@attribute(local_name='finalDefault', enum=['#all', 'extension', 'restriction', 'list', 'union'])
@attribute(local_name='blockDefault', enum=['#all', 'extension', 'restriction', 'substitution'])
@attribute(local_name='attributeFormDefault', enum=FORM_CHOICE_ENUMERATION, default='unqualified')
@attribute(local_name='elementFormDefault', enum=FORM_CHOICE_ENUMERATION, default='unqualified')
@attribute(local_name='id', type=IdType)
@element(local_name='include', list='tags', cls=IncludeElement, min=0, max=None)
@element(local_name='import', list='tags', cls=ImportElement, min=0, max=None)
@element(local_name='redefine', list='tags', cls=RedefineElement, min=0, max=None)
@element(local_name='annotation', list='tags', cls=AnnotationElement, min=0, max=None)
@element(local_name='simpleType', list='tags', cls=SimpleTypeType, min=0, max=None)
@element(local_name='complexType', list='tags', cls=ComplexTypeType, min=0, max=None)
@element(local_name='group', list='tags', cls=GroupType, min=0, max=None)
@element(local_name='attributeGroup', list='tags', cls=AttributeGroupType, min=0, max=None)
@element(local_name='element', list='tags', cls=ElementType, min=0, max=None)
@element(local_name='attribute', list='tags', cls=AttributeType, min=0, max=None)
@element(local_name='notation', list='tags', cls=NotationElement, min=0, max=None)
class SchemaElement(AnyTypeType):
    def add_enumeration(self, name, enum):
        self._enumerations[name] = enum

    def add_tag_mapping(self, name, class_name):
        self._tag_mapping[name] = class_name

    def stub(self, path):
        self._enumerations = {}
        self._tag_mapping = {}

        for c in self.tags:
            if c.tag_name in ['simpleType', 'complexType', 'element']:
                c.stub(path, self)

        with open(os.path.join(path, '__init__.py'), 'w') as f:
            f.write(STUB_HEADER)
            f.write('ELEMENT_MAP = {\n')
            for name in sorted(self._tag_mapping.keys()):
                f.write("    '{" + self.targetNamespace + '}' + name + "': '" + self._tag_mapping[name] + "',\n")
            f.write('}\n\n')

        logger.debug('Wrote __init__.py')



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import functools
import importlib
import logging
import os
import sys
import types

from .AttributeMapper import AttributeMapper
from .ContentMapper import ContentMapper
from .ElementMapper import ElementMapper

logger = logging.getLogger(__name__)

def attribute(*args, **kwargs):
    '''
    Decorator to map xml elements to model children

    Used as follows::

        @attribute(local_name='test', type=StringType)
        class ModelWithTestAttribute(Model):
            pass

    :param str namespace: The xml namespace to match. It can also be * to match any namespace. If not specified, it defaults to the parent element's namespace.
    :param str local_name: Required. The local name of the xml attribute we're matching. It can also be * to match any local name.
    :param str into: The python attribute to store the value of the attribute into. Defaults to the local_name if not specified.
    :param type: The type of the expected value. Types are stored directly as data, no enclosed in a model class. Types usually restrict the domain values.
    :type type: class or tuple(package_str, class_str)
    :param enum: Enumeration the attribute's value must be from
    :type enum: list or tuple
    :param str pattern: Pattern which the value of the attribute must match.
    :param bool required: True or False (default). Specifies if the attribute is required.
    :param default: The default value of an attribute if it isn't specified. The (implicit) default is None or the first item of an *enum*.
    :param min: The minimum value of the attribute. Can be numeric or None (the default).
    :param max: The maximum value of the attribute. Can be numeric or None (the default).
    :param prohibited: The attribute should not appear in the element.
    '''
    def wrapper(cls):
        functools.update_wrapper(wrapper, cls)
        cls._add_attribute_mapper(AttributeMapper(**kwargs))

        return cls
    return wrapper

def element(*args, **kwargs):
    '''
    Decorator to map xml elements to model children.

    One of *type* or *cls* must be specified. If the type/class cannot be passed in
    by object, a tuple with the package and name of the class may be passed to
    defer the import of the class.

    Used as follows::

        @element(namespace='http://jaymes.biz/test', local_name='test',
            list='test', type=StringType)
        class ModelWithTestChildrenInAList(Model):
            pass

    :param namespace: The xml namespace to match. It can also be * to match any namespace. If not specified, it defaults to the parent element.
    :param local_name: Required. The local name of the xml element we're matching. It can also be * to match any local name.
    :param into: The python attribute to store the value of the element into. Defaults to the local_name if not specified.
    :param list: The python attribute to store the value of the element into (as a list).  Defaults to the local_name if not specified.
    :param dict: The python attribute to store the value of the element into (as a dict). Defaults to the local_name if not specified.
    :param dict_key: The attribute of the sub-element to use as the key of the dict. By default it is the *id* attribute.
    :param dict_value: The attribute of the sub-element to use as the value of the dict. By default it is the value of the element.
    :param type: The type of the expected value. Types are stored directly as data, no enclosed in a model class. Types usually restrict the domain values.
    :param cls: The model class with which to load the element.
    :param min: The minimum number of elements to be present. Can be numeric or None (the  default).
    :param max: The maximum number of elements to be present. Can be numeric or None (the default).
    :param enum: Enumeration to which the value of the element must belong.
    :param pattern: Pattern which the value of the element must match.
    :param nillable: If True, the element can be nil (from the xsi spec). False specifies that it cannot (the default).
    '''
    def wrapper(cls):
        functools.update_wrapper(wrapper, cls)
        cls._add_element_mapper(ElementMapper(**kwargs))

        return cls
    return wrapper

def content(*args, **kwargs):
    '''
    Decorator to map xml element content to model data.

    Used as follows::

        @content(enum=('value1', 'value2'))
        class Value1Or2(Model):
            pass

    :param enum: Enumeration the attribute's value must be from
    :param pattern: Pattern which the value of the attribute must match.
    :param type: Type against which a value must validate
    :param min: The minimum value of the attribute. Can be numeric or None (the default).
    :param max: The maximum value of the attribute. Can be numeric or None (the default).
    '''
    def wrapper(cls):
        functools.update_wrapper(wrapper, cls)
        cls._add_content_mapper(ContentMapper(**kwargs))

        return cls
    return wrapper



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .exceptions import *
from .Node import Node

logger = logging.getLogger(__name__)

class Namespace(Node):
    '''
    Class representing a XML namespace
    '''
    def __init__(self, prefix, uri, parent=None):
        super().__init__(parent=parent)

        self.prefix = prefix
        self.uri = uri

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)

    def get_type(self):
        '''
        Return the type of the node
        '''
        return 'namespace'

    def get_string_value(self):
        '''
        Return the string value of the node
        '''
        return self.uri

    def get_expanded_name(self):
        '''
        Return the expanded name of the node
        '''
        return (None, self.prefix)

    def get_document_order(self):
        if self._parent is None:
            raise UnattachedElementException('Element ' + str(self) + ' is not attached to a document')

        do = self._parent.get_document_order()
        ordered_ns = [self._parent.namespace_nodes[k] for k in self._parent.namespace_nodes.keys()]
        return do + ordered_ns.index(self)



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .Axis import Axis
from ..exceptions import XPathSyntaxException

logger = logging.getLogger(__name__)
class Step(object):
    def __init__(self):
        self.children = []

    def evaluate(self, context_node, context_position, context_size, variables):
        from ..Document import Document
        if len(self.children) == 1:
            logger.debug('Collecting nodes with ' + str(self.children[0]) + ' for context node ' + str(context_node))
            ns = self.children[0].evaluate(context_node, context_position, context_size, variables)
            logger.debug('Nodes from ' + str(self.children[0]) + ': [' + ','.join([str(x) for x in ns]) + ']')

            logger.debug(str(self) + ' nodeset: [' + ','.join([str(x) for x in ns]) + ']')
            return ns
        elif len(self.children) == 2:
            logger.debug('Collecting context nodes with ' + str(self.children[0]) + ' for context node ' + str(context_node))
            context_nodes = self.children[0].evaluate(context_node, context_position, context_size, variables)
            logger.debug('Context nodes from ' + str(self.children[0]) + ': [' + ','.join([str(x) for x in context_nodes]) + ']')

            ns = []
            for i, cn in enumerate(context_nodes):
                logger.debug('Evaluating ' + str(self.children[1]) + ' with context ' + str(cn))
                ns.extend(self.children[1].evaluate(cn, i+1, len(context_nodes), variables))

            ns = Document.order_sort(ns)
            logger.debug(str(self) + ' nodeset: [' + ','.join([str(x) for x in ns]) + ']')
            return ns
        else:
            raise XPathSyntaxException('Steps require between 1 and 2 children')

    def __str__(self):
        return 'Step ' + hex(id(self)) + ': [' + ','.join([str(x) for x in self.children]) + ']'



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os.path

from . import FORM_CHOICE_ENUMERATION
from ..decorators import *
from .AllNniType import AllNniType
from .AnnotatedType import AnnotatedType
from .BooleanType import BooleanType
from .ComplexTypeType import ComplexTypeType
from .KeybaseType import KeybaseType
from .KeyRefElement import KeyRefElement
from .NCNameType import NCNameType
from .NonNegativeIntegerType import NonNegativeIntegerType
from .QNameType import QNameType
from .SimpleTypeType import SimpleTypeType
from .StringType import StringType

logger = logging.getLogger(__name__)

@element(local_name='simpleType', list='tags', cls=SimpleTypeType, min=0)
@element(local_name='complexType', list='tags', cls=ComplexTypeType, min=0)
@element(local_name='unique', list='tags', cls=KeybaseType, min=0, max=None)
@element(local_name='key', list='tags', cls=KeybaseType, min=0, max=None)
@element(local_name='keyref', list='tags', cls=KeyRefElement, min=0, max=None)
@attribute(local_name='name', type=NCNameType)
@attribute(local_name='ref', type=QNameType)
@attribute(local_name='type', type=QNameType)
@attribute(local_name='substitutionGroup', type=QNameType)
@attribute(local_name='default', type=StringType)
@attribute(local_name='fixed', type=StringType)
@attribute(local_name='nillable', type=BooleanType, default=False)
@attribute(local_name='abstract', type=BooleanType, default=False)
@attribute(local_name='final', enum=['#all', 'extension', 'restriction'])
@attribute(local_name='block', enum=['#all', 'extension', 'restriction', 'substitution'])
@attribute(local_name='form', enum=FORM_CHOICE_ENUMERATION)
@attribute(local_name='minOccurs', type=NonNegativeIntegerType, default=1)
@attribute(local_name='maxOccurs', type=AllNniType, default=1)
@attribute(local_name='*')
class ElementType(AnnotatedType):
    def stub(self, path, schema):
        class_name = ''.join(cap_first(w) for w in self.name.split('_'))
        if not class_name.endswith('Element'):
            class_name = class_name + 'Element'
        schema.add_tag_mapping(self.name, class_name)

        super().stub(path, schema, class_name)

    def get_defs(self, schema, top_level):
        if self.ref is not None:
            return schema.find_reference(self.ref).get_defs(schema, top_level)

        model_map = {'elements': [], 'attributes': {}}

        e = {}

        e['min'] = self.minOccurs

        if self.maxOccurs == 'unbounded':
            e['max'] = None
        else:
            e['max'] = self.maxOccurs

        if self.default is not None:
            e['default'] = self.default

        if self.nillable:
            e['nillable'] = True

        # TODO type vs. class detection
        e['class'] = self.type

        if self.maxOccurs == 'unbounded':
            e['list'] = self.name + 's'

        e['tag_name'] = self.name

        model_map['elements'].append(e)
        logger.debug('Adding element ' + str(e))

        for t in self.tags:
            defs = t.get_defs(schema, top_level)
            model_map['elements'].extend(defs['elements'])
            model_map['attributes'].update(defs['attributes'])
        return model_map



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

XML_SPACE_ENUMERATION = (
    'default',
    # The value "default" signals that applications' default white-space
    # processing modes are acceptable for this element
    'preserve',
    # the value "preserve" indicates the intent that applications preserve all
    # the white space
)
''' Enumeration of the possible whitespace processing modes '''

from .exceptions import *
from .Attribute import Attribute
from .CharacterData import CharacterData
from .Comment import Comment
from .Document import Document
from .Element import Element
from .Namespace import Namespace
from .ProcessingInstruction import ProcessingInstruction



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

import expatriate

from ..Node import Node
from .exceptions import *
from .Mapper import Mapper

logger = logging.getLogger(__name__)

class ContentMapper(Mapper):
    '''
        **kwargs**

        enum
            Enumeration the attribute's value must be from
        pattern
            Pattern which the value must match.
        type
            Type against which a value must validate

        min
            The minimum value of the content. Can be numeric or None (the
            default).
        max
            The maximum value of the content. Can be numeric or None (the
            default).
    '''

    def initialize(self, model):
        from .Model import Model
        pass

    def validate(self, model):
        from .Model import Model
        pass

    def produce_in(self, el, model, id_):
        from .Model import Model
        logger.debug(str(self) + ' producing ' + str(id_) + ' in ' + str(el))
        el.children.append(expatriate.CharacterData(str(id_)))



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

ELEMENT_MAP = {
    ('http://www.w3.org/2001/XMLSchema', 'schema'): 'SchemaElement',
}

i_ = r'[A-Z_a-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD]'
c_ = r'[-.0-9A-Z_a-z\u00B7\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u037D\u037F-\u1FFF\u200C-\u200D\u203F\u2040\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD]'

FORM_CHOICE_ENUMERATION = [
    'qualified',
    'unqualified',
]

def cap_first(s):
    return s[0].upper() + s[1:]



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .exceptions import *
from .Node import Node
from .xpath.Literal import Literal

logger = logging.getLogger(__name__)

class Attribute(Node):
    '''
    Class representing a XML attribute node
    '''
    def __init__(self, local_name, value, parent=None, prefix=None, namespace=None):
        super().__init__(parent=parent)

        self._prefix = prefix
        self._local_name = local_name
        self._namespace = namespace

        self.value = value

    @property
    def name(self):
        if self._prefix is None:
            return self._local_name
        else:
            return self._prefix + ':' + self._local_name

    @name.setter
    def name(self, name):
        if ':' in name:
            self._prefix, colon, self._name = name.partition(':')
        else:
            self._prefix = None
            self._name = name

        self._namespace = self.prefix_to_namespace(self._prefix)

    @property
    def local_name(self):
        return self._local_name

    @local_name.setter
    def local_name(self, local_name):
        self._local_name = local_name

    @property
    def prefix(self):
        return self._prefix

    @prefix.setter
    def prefix(self, prefix):
        self._prefix = prefix
        self._namespace = self.prefix_to_namespace(self._prefix)

    @property
    def namespace(self):
        return self._namespace

    @namespace.setter
    def namespace(self, namespace):
        self._namespace = namespace
        self._prefix = self.namespace_to_prefix(namespace)

    def get_type(self):
        '''
        Return the type of the node
        '''
        return 'attribute'

    def get_string_value(self):
        '''
        Return the string value of the node
        '''
        return self.value

    def get_expanded_name(self):
        '''
        Return the expanded name of the node
        '''
        return (self.namespace, self.local_name)

    def __eq__(self, other):
        if isinstance(other, Literal):
            return self.value == other.value
        elif isinstance(other, str):
            return self.value == other
        elif isinstance(other, int) or isinstance(other, float):
            return self.value == str(other)
        return object.__eq__(self, other)

    def __str__(self):
        s = self.__class__.__name__ + ' ' + hex(id(self)) + ' '
        if self.namespace is not None:
            s += self.namespace + ':'
        s += self.local_name + '=' + self.value
        return s

    def get_document_order(self):
        if self._parent is None:
            raise UnattachedElementException('Element ' + str(self) + ' is not attached to a document')

        do = self._parent.get_document_order()
        do += len(self._parent.namespace_nodes)
        ordered_attr = [self._parent.attribute_nodes[k] for k in sorted(self._parent.attribute_nodes.keys())]
        return do + ordered_attr.index(self)



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import datetime as dt  # avoids conflict with init kwarg
import logging
import re

from ..decorators import *

logger = logging.getLogger(__name__)

class SevenPropertyModel(object):
    def __init__(
        self,
        datetime=None,
        time=None,
        date=None,
        year=None,
        month=None,
        day=None,
        hour=None,
        minute=None,
        second=None,
        timezoneOffset=None
    ):

        self._year = None
        self._month = None
        self._day = None
        self._hour = None
        self._minute = None
        self._second = None
        self._timezoneOffset = None

        if datetime is not None:
            year = datetime.year
            month = datetime.month
            day = datetime.day
            hour = datetime.hour
            minute = datetime.minute
            second = datetime.second
            if datetime.tzinfo is not None:
                try:
                    timezoneOffset = datetime.tzinfo.utcoffset(datetime)
                    if isinstance(timezoneOffset, dt.timedelta):
                        if timezoneOffset.days > 0:
                            day += timezoneOffset.days
                        timezoneOffset = timezoneOffset.seconds / 60
                except:
                    pass

        if time is not None:
            hour = time.hour
            minute = time.minute
            second = time.second
            if time.tzinfo is not None:
                try:
                    timezoneOffset = time.tzinfo.utcoffset(time)
                    if isinstance(timezoneOffset, dt.timedelta):
                        if timezoneOffset.days > 0:
                            day += timezoneOffset.days
                        timezoneOffset = timezoneOffset.seconds / 60
                except:
                    pass

        if date is not None:
            year = date.year
            month = date.month
            day = date.day

        if year is not None:
            self.year = year

        if month is not None:
            self.month = month

        if day is not None:
            self.day = day

        if hour is not None:
            self.hour = hour

        if minute is not None:
            self.minute = minute

        if second is not None:
            self.second = second

        if timezoneOffset is not None:
            self.timezoneOffset = timezoneOffset

    @property
    def year(self):
        return self._year

    @year.setter
    def year(self, value):
        if isinstance(value, str):
            value = int(value)
        elif not isinstance(value, int):
            raise TypeError('Year must be an int')

        self._year = value

    @property
    def month(self):
        return self._month

    @month.setter
    def month(self, value):
        if isinstance(value, str):
            value = int(value)
        elif not isinstance(value, int):
            raise TypeError('month must be an int')

        if value < 1 or value > 12:
            raise ValueError('month must be between 1 and 12')

        self._month = value

    @property
    def day(self):
        return self._day

    @day.setter
    def day(self, value):
        if isinstance(value, str):
            value = int(value)
        elif not isinstance(value, int):
            raise TypeError('day must be an int')

        if value < 1 or value > 31:
            raise ValueError('day must be between 1 and 31')

        self._day = value

    @property
    def hour(self):
        return self._hour

    @hour.setter
    def hour(self, value):
        if isinstance(value, str):
            value = int(value)
        elif not isinstance(value, int):
            raise TypeError('hour must be an int')

        if value < 0 or value > 23:
            raise ValueError('hour must be between 0 and 23')

        self._hour = value

    @property
    def minute(self):
        return self._minute

    @minute.setter
    def minute(self, value):
        if isinstance(value, str):
            value = int(value)
        elif not isinstance(value, int):
            raise TypeError('minute must be an int')

        if value < 0 or value > 59:
            raise ValueError('minute must be between 0 and 59')

        self._minute = value

    @property
    def second(self):
        return self._second

    @second.setter
    def second(self, value):
        if isinstance(value, str) or isinstance(value, int):
            value = float(value)
        elif not isinstance(value, float):
            raise TypeError('second must be a float')

        if value < 0.0 or value >= 60.0:
            raise ValueError('second must be between 0 and 60')

        self._second = value

    @property
    def timezoneOffset(self):
        return self._timezoneOffset

    @timezoneOffset.setter
    def timezoneOffset(self, value):
        if isinstance(value, str):
            if value == 'Z' or value == '+00:00' or value == '-00:00':
                value = 0
            elif ':' in value:
                tz_hour, tz_minute = value.split(':')
                value = (int(tz_hour) * 60) + int(tz_minute)
            else:
                value = int(timezoneOffset)
        elif not isinstance(value, int):
            raise TypeError('timezoneOffset must be an int')

        if value < -840 or value > 840:
            raise ValueError('timezoneOffset must be between -840 and 840')

        self._timezoneOffset = value

    def _tz_to_string(self):
        if self._timezoneOffset is None:
            return ''
        elif self._timezoneOffset == 0:
            return 'Z'
        else:
            tz_hour, tz_minute = divmod(self._timezoneOffset, 60)
            if self._timezoneOffset > 0:
                return '+%02d:%02d' % (int(tz_hour), tz_minute)
            else:
                return '-%02d:%02d' % (int(tz_hour), tz_minute)

    def __str__(self):
        if (
            self._year is None
            and self._month is None
            and self._day is not None
            and self._hour is None
            and self._minute is None
            and self._second is None
        ):
            # gDay
            return "---%02d%s" % (self._day, self._tz_to_string())

        elif (
            self._year is None
            and self._month is not None
            and self._day is None
            and self._hour is None
            and self._minute is None
            and self._second is None
        ):
            # gMonth
            return "--%02d%s" % (self._month, self._tz_to_string())

        elif (
            self._year is None
            and self._month is not None
            and self._day is not None
            and self._hour is None
            and self._minute is None
            and self._second is None
        ):
            # gMonthDay
            return "--%02d-%02d%s" % (self._month, self._day, self._tz_to_string())

        elif (
            self._year is not None
            and self._month is None
            and self._day is None
            and self._hour is None
            and self._minute is None
            and self._second is None
        ):
            # gYear
            return "%04d%s" % (self._year, self._tz_to_string())

        elif (
            self._year is not None
            and self._month is not None
            and self._day is None
            and self._hour is None
            and self._minute is None
            and self._second is None
        ):
            # gYearMonth
            return "%04d-%02d%s" % (self._year, self._month, self._tz_to_string())

        elif (
            self._year is not None
            and self._month is not None
            and self._day is not None
            and self._hour is None
            and self._minute is None
            and self._second is None
        ):
            # Date
            return "%04d-%02d-%02d%s" % (self._year, self._month, self._day, self._tz_to_string())

        elif (
            self._year is None
            and self._month is None
            and self._day is None
            and self._hour is not None
            and self._minute is not None
            and self._second is not None
        ):
            # Time
            if self._second - int(self._second) == 0.0:
                return "%02d:%02d:%02.0f%s" % (self._hour, self._minute, self._second, self._tz_to_string())
            else:
                return "%02d:%02d:%02.6f%s" % (self._hour, self._minute, self._second, self._tz_to_string())

        elif (
            self._year is not None
            and self._month is not None
            and self._day is not None
            and self._hour is not None
            and self._minute is not None
            and self._second is not None
        ):
            # DateTime
            if self._second - int(self._second) == 0.0:
                return "%04d-%02d-%02dT%02d:%02d:%02.0f%s" % (
                    self._year,
                    self._month,
                    self._day,
                    self._hour,
                    self._minute,
                    self._second,
                    self._tz_to_string()
                )
            else:
                return "%04d-%02d-%02dT%02d:%02d:%02.6f%s" % (
                    self._year,
                    self._month,
                    self._day,
                    self._hour,
                    self._minute,
                    self._second,
                    self._tz_to_string()
                )

        else:
            raise NotImplementedError(
                'Value set '
                + str([
                    self._year,
                    self._month,
                    self._day,
                    self._hour,
                    self._minute,
                    self._second
                ])
                + ' cannot be converted to a string'
            )

    def __eq__(self, other):
        if not isinstance(other, SevenPropertyModel):
            return False

        if self._year != other._year:
            return False

        if self._month != other._month:
            return False

        if self._day != other._day:
            return False

        if self._hour != other._hour:
            return False

        if self._minute != other._minute:
            return False

        if self._second != other._second:
            return False

        if self._timezoneOffset != other._timezoneOffset:
            return False

        return True

    def to_datetime(self):
        ms = int((self._second - int(self._second)) * 1000)

        if self._timezoneOffset is None:
            return dt.datetime(
                year=self._year,
                month=self._month,
                day=self._day,
                hour=self._hour,
                minute=self._minute,
                second=int(self._second),
                microsecond=ms
            )
        else:
            tz = dt.timezone(
                offset=dt.timedelta(
                    minutes=self._timezoneOffset
                )
            )

            return dt.datetime(
                year=self._year,
                month=self._month,
                day=self._day,
                hour=self._hour,
                minute=self._minute,
                second=int(self._second),
                microsecond=ms,
                tzinfo=tz
            )

    def to_date(self):
        # TODO work around date's timezone naiveté
        return dt.date(
            year=self._year,
            month=self._month,
            day=self._day
        )

    def to_time(self):
        ms = int((self._second - int(self._second)) * 1000)

        if self._timezoneOffset is None:
            return dt.time(
                hour=self._hour,
                minute=self._minute,
                second=int(self._second),
                microsecond=ms
            )
        else:
            tz = dt.timezone(
                offset=dt.timedelta(
                    minutes=self._timezoneOffset
                )
            )

            return dt.time(
                hour=self._hour,
                minute=self._minute,
                second=int(self._second),
                microsecond=ms,
                tzinfo=tz
            )



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from publishsubscribe import PublishingDict, Subscriber

from .Attribute import Attribute
from .exceptions import *
from .Namespace import Namespace
from .Node import Node
from .Parent import Parent

logger = logging.getLogger(__name__)

class Element(Parent, Subscriber):
    '''
    Class representing a XML element
    '''
    def __init__(self, local_name, attributes=None, prefix=None, namespace=None, parent=None):
        super().__init__(parent=parent)

        if attributes is None:
            self._attributes = PublishingDict({})
        else:
            self._attributes = PublishingDict(attributes)
        self._attributes.subscribe(self)

        self._prefix = prefix
        self._namespace = namespace
        self._local_name = local_name

        self._init_namespaces()
        self._init_attributes()

        if namespace is not None and namespace not in self._namespace_to_prefixes:
            logger.debug(str(self) + ' parent does not define namespace ' + namespace
                + '; adding to attributes')
            if prefix is None:
                self.attributes['xmlns'] = namespace
            else:
                self.attributes['xmlns:' + prefix] = namespace

    # redefine parent property
    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        self._parent = parent
        self._init_namespaces()

    @property
    def attributes(self):
        return self._attributes

    @property
    def name(self):
        if self._prefix is None:
            return self._local_name
        else:
            return self._prefix + ':' + self._local_name

    @name.setter
    def name(self, name):
        if ':' in name:
            self._prefix, colon, self._name = name.partition(':')
        else:
            self._prefix = None
            self._name = name

        if self._parent is not None:
            self._namespace = self.prefix_to_namespace(self._prefix)

    @property
    def local_name(self):
        return self._local_name

    @local_name.setter
    def local_name(self, local_name):
        self._local_name = local_name

    @property
    def prefix(self):
        return self._prefix

    @prefix.setter
    def prefix(self, prefix):
        self._prefix = prefix
        if self._parent is not None:
            self._namespace = self.prefix_to_namespace(self._prefix)

    @property
    def namespace(self):
        return self._namespace

    @namespace.setter
    def namespace(self, namespace):
        self._namespace = namespace
        if self._parent is not None:
            self._prefix = self.namespace_to_prefix(namespace)

    def data_added(self, publisher, id_, item):
        logger.debug(str(self) + ' added attributes: ' + str(id_))
        self._init_attributes()
        if id_.startswith('xmlns'):
            self._init_namespaces()

    def data_updated(self, publisher, id_, old_item, new_item):
        logger.debug(str(self) + ' updated attributes: ' + str(id_))
        self._init_attributes()
        if id_.startswith('xmlns'):
            self._init_namespaces()

    def data_deleted(self, publisher, id_, item):
        logger.debug(str(self) + ' deleted attributes: ' + str(id_))
        self._init_attributes()
        if id_.startswith('xmlns'):
            self._init_namespaces()

    def _init_attributes(self):
        # create nodes for each of the attributes
        self.attribute_nodes = {}
        for k, v in self._attributes.items():
            if ':' in k:
                prefix, colon, local_name = k.partition(':')
                # check prefix
                namespace = self.prefix_to_namespace(prefix)
            else:
                local_name = k
                prefix = None
                namespace = None
            n = Attribute(local_name, v, parent=self, prefix=prefix, namespace=namespace)
            self.attribute_nodes[k] = n

    def _init_namespaces(self):
        if isinstance(self._parent, Element):
            self._prefix_to_namespace = self._parent._prefix_to_namespace.copy()
            self._namespace_to_prefixes = self._parent._namespace_to_prefixes.copy()
        else:
            # parent is-a Document or None
            self._prefix_to_namespace = {
                'xml': 'http://www.w3.org/XML/1998/namespace',
            }
            self._namespace_to_prefixes = {v: k for k, v in self._prefix_to_namespace.items()}

        # check for prefix namespaces
        for k, v in self._attributes.items():
            if k.startswith('xmlns:'):
                prefix = k.partition(':')[2]
                if prefix in self._prefix_to_namespace:
                    raise PrefixRedefineException('Prefix ' + prefix
                        + ' has already been used but is being redefined')
                self._prefix_to_namespace[prefix] = v
                self._namespace_to_prefixes[v] = prefix
                logger.debug(str(self) + ' Added prefix ' + prefix + ' for uri ' + v)
            elif k.startswith('xmlns'):
                self._prefix_to_namespace[None] = v
                self._namespace_to_prefixes[v] = None
                logger.debug(str(self) + ' Added prefix None for uri ' + v)

        # now that we've parsed the namespace attributes, we can figure out missing info
        if self._namespace is None and self._prefix is None and None in self._prefix_to_namespace:
            self._namespace = self._prefix_to_namespace[None]
        elif self._namespace is None and self._prefix is not None:
            self._namespace = self.prefix_to_namespace(self._prefix)

        # create nodes for each of the namespaces
        self.namespace_nodes = {}
        for prefix in self._prefix_to_namespace.keys():
            uri = self._prefix_to_namespace[prefix]
            n = Namespace(prefix, uri, parent=self)
            self.namespace_nodes[prefix] = n

    def prefix_to_namespace(self, prefix):
        logger.debug(str(self) + ' resolving prefix: ' + str(prefix)
            + ' using ' + str(self._prefix_to_namespace))

        if prefix in self._prefix_to_namespace:
            return self._prefix_to_namespace[prefix]
        else:
            return super().prefix_to_namespace(prefix)

    def namespace_to_prefix(self, namespace):
        logger.debug(str(self) + ' resolving namespace: ' + str(namespace)
            + ' using ' + str(self._namespace_to_prefixes))

        if namespace in self._namespace_to_prefixes:
            return self._namespace_to_prefixes[namespace]
        else:
            return super().namespace_to_prefix(namespace)

    def escape_attribute(self, text):
        return self.escape(text).replace('"', '&quot;')

    def produce(self):
        logger.debug(str(self) + ' producing xml: ' + self.name + ' attributes '
            + str(self.attributes) + '; ' + str(len(self.children))
            + ' children')

        s = '<' + self.name
        for k, v in self.attributes.items():
            s += ' ' + self.escape(k) + '="' + self.escape_attribute(v) + '"'

        if len(self.children) == 0:
            s += '/>'
        else:
            s += '>'
            for c in self.children:
                s += c.produce()
            s += '</' + self.name + '>'

        return s

    def get_type(self):
        '''
        Return the type of the node
        '''
        return 'element'

    def get_string_value(self):
        '''
        Return the string value of the node
        '''
        from .CharacterData import CharacterData
        s = ''
        for c in self.children:
            if isinstance(c, CharacterData):
                s += c.data
            elif isinstance(c, Element):
                s += c.get_string_value()
        return s

    def get_expanded_name(self):
        '''
        Return the expanded name of the node
        '''
        return (self.namespace, self.local_name)

    def __str__(self):
        s = self.__class__.__name__ + ' ' + hex(id(self)) + ' ' + self.name
        if 'id' in self.attributes:
            s += ' id=' + self.attributes['id']
        if 'name' in self.attributes:
            s += ' name=' + self.attributes['name']
        return s

    def find_by_id(self, id_):
        logger.debug(str(self) + ' checking attributes for id: ' + str(id_))
        for k, v in self.attributes.items():
            k = k.lower()
            if k.endswith(':id') or k == 'id':
                logger.debug(str(self) + ' found id: ' + str(v))
                if v == id_:
                    logger.debug(str(self) + ' matches id: ' + str(id_))
                    return self
                else:
                    logger.debug(str(self) + ' id ' + str(v) + ' does not match id: ' + str(id_))

        return super().find_by_id(id_)

    def get_node_count(self):
        do = 1 + len(self.namespace_nodes) + len(self.attribute_nodes)
        for c in self.children:
            do += c.get_node_count()
        return do

    def is_nil(self):
        try:
            prefix = self.namespace_to_prefix('http://www.w3.org/2001/XMLSchema-instance')
        except UnknownNamespaceException:
            # if the namespace hasn't been defined, attr doesn't exist either
            return False

        name = prefix + ':nil'
        if name in self.attributes and self.attributes[name] == 'true':
            return True
        else:
            return False



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from ..decorators import *
from .IntegerType import IntegerType

logger = logging.getLogger(__name__)

class NonPositiveIntegerType(IntegerType):
    def parse_value(self, value):
        value = super().parse_value(value)

        if value > 0:
            raise ValueError('xs:nonPositiveInteger cannot be > 0')

        return value



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re

from ..decorators import *
from .List import List
from .TokenType import TokenType

logger = logging.getLogger(__name__)

class NamespaceListType(TokenType):
    def parse_item(self, item_value):
        if (
            item_value == '##any'
            or item_value == '##other'
            or item_value == '##targetNamespace'
            or item_value == '##local'
        ):
            return item_value
        return Token().parse_value(item_value)

    def produce_item(self, item_value):
        if (
            item_value == '##any'
            or item_value == '##other'
            or item_value == '##targetNamespace'
            or item_value == '##local'
        ):
            return item_value
        return Token().produce_value(item_value)



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging

from . import PROPERTY_NAME_ENUMERATION
from ..decorators import *
from ..Model import Model
from ..xs.NormalizedStringType import NormalizedStringType

logger = logging.getLogger(__name__)

@attribute(local_name='name', required=True, enum=PROPERTY_NAME_ENUMERATION)
@attribute(local_name='value', required=True, type=NormalizedStringType)
class HasPropertyElement(Model):
    pass
